@@ -20,8 +20,13 @@ use core::{
 	iter::FromIterator,
 	marker::PhantomData,
 	mem,
+	mem::{
+		MaybeUninit,
+		forget,
+	},
 	ops::{Deref, Range, RangeInclusive},
 	time::Duration,
+	ptr,
 };
 use core::num::{
 	NonZeroI8,
@@ -35,7 +40,6 @@ use core::num::{
 	NonZeroU64,
 	NonZeroU128,
 };
-use arrayvec::ArrayVec;
 
 use byte_slice_cast::{AsByteSlice, AsMutByteSlice, ToMutByteSlice};
 
@@ -367,6 +371,23 @@ mod feature_full_wrapper_type_encode {
 	impl EncodeLike<String> for &str {}
 }
 
+#[cfg(feature = "bytes")]
+mod feature_wrapper_bytes {
+	use super::*;
+	use bytes::Bytes;
+
+	impl WrapperTypeEncode for Bytes {}
+	impl EncodeLike for Bytes {}
+	impl EncodeLike<&[u8]> for Bytes {}
+	impl EncodeLike<Vec<u8>> for Bytes {}
+	impl EncodeLike<Bytes> for &[u8] {}
+	impl EncodeLike<Bytes> for Vec<u8> {}
+
+	impl WrapperTypeDecode for Bytes {
+		type Wrapped = Vec<u8>;
+	}
+}
+
 impl<T, X> Encode for X where
 	T: Encode + ?Sized,
 	X: WrapperTypeEncode<Target = T>,
@@ -637,7 +658,91 @@ pub(crate) fn encode_slice_no_len<T: Encode, W: Output + ?Sized>(slice: &[T], de
 	}
 }
 
-/// Decode the slice (without prepended the len).
+/// Decode the array.
+///
+/// This is equivalent to decoding all the element one by one, but it is optimized for some types.
+#[inline]
+pub(crate) fn decode_array<I: Input, T: Decode, const N: usize>(input: &mut I) -> Result<[T; N], Error> {
+	#[inline]
+	fn general_array_decode<I: Input, T: Decode, const N: usize>(input: &mut I) -> Result<[T; N], Error> {
+		let mut uninit = <MaybeUninit<[T; N]>>::uninit();
+		// The following line coerces the pointer to the array to a pointer
+		// to the first array element which is equivalent.
+		let mut ptr = uninit.as_mut_ptr() as *mut T;
+		for _ in 0..N {
+			let decoded = T::decode(input)?;
+			// SAFETY: We do not read uninitialized array contents
+			//		 while initializing them.
+			unsafe {
+				ptr::write(ptr, decoded);
+			}
+			// SAFETY: Point to the next element after every iteration.
+			//		 We do this N times therefore this is safe.
+			ptr = unsafe { ptr.add(1) };
+		}
+		// SAFETY: All array elements have been initialized above.
+		let init = unsafe { uninit.assume_init() };
+		Ok(init)
+	}
+
+	// Description for the code below.
+	// It is not possible to transmute `[u8; N]` into `[T; N]` due to this issue:
+	// 		https://github.com/rust-lang/rust/issues/61956
+	//
+	// Workaround: Transmute `&[u8; N]` into `&[T; N]` and interpret that reference as value.
+	// ```
+	// let mut array: [u8; N] = [0; N];
+	// let ref_typed: &[T; N] = unsafe { mem::transmute(&array) };
+	// let typed: [T; N] = unsafe { ptr::read(ref_typed) };
+	// forget(array);
+	// Here `array` and `typed` points on the same memory.
+	// Function returns `typed` -> it is not dropped, but `array` will be dropped.
+	// To avoid that `array` should be forgotten.
+	// ```
+	macro_rules! decode {
+		( u8 ) => {{
+			let mut array: [u8; N] = [0; N];
+			input.read(&mut array[..])?;
+			let ref_typed: &[T; N] = unsafe { mem::transmute(&array) };
+			let typed: [T; N] = unsafe { ptr::read(ref_typed) };
+			forget(array);
+			Ok(typed)
+		}};
+		( i8 ) => {{
+			let mut array: [i8; N] = [0; N];
+			let bytes = unsafe { mem::transmute::<&mut [i8], &mut [u8]>(&mut array[..]) };
+			input.read(bytes)?;
+
+			let ref_typed: &[T; N] = unsafe { mem::transmute(&array) };
+			let typed: [T; N] = unsafe { ptr::read(ref_typed) };
+			forget(array);
+			Ok(typed)
+		}};
+		( $ty:ty ) => {{
+			if cfg!(target_endian = "little") {
+				let mut array: [$ty; N] = [0; N];
+				let bytes = <[$ty] as AsMutByteSlice<$ty>>::as_mut_byte_slice(&mut array[..]);
+				input.read(bytes)?;
+				let ref_typed: &[T; N] = unsafe { mem::transmute(&array) };
+				let typed: [T; N] = unsafe { ptr::read(ref_typed) };
+				forget(array);
+				Ok(typed)
+			} else {
+				general_array_decode(input)
+			}
+		}};
+	}
+
+	with_type_info! {
+		<T as Decode>::TYPE_INFO,
+		decode,
+		{
+			general_array_decode(input)
+		},
+	}
+}
+
+/// Decode the vec (without prepended the len).
 ///
 /// This is equivalent to decode all elements one by one, but it is optimized in some
 /// situation.
@@ -706,16 +811,9 @@ impl<T: Encode, const N: usize> Encode for [T; N] {
 }
 
 impl<T: Decode, const N: usize> Decode for [T; N] {
+	#[inline]
 	fn decode<I: Input>(input: &mut I) -> Result<Self, Error> {
-		let mut array = ArrayVec::new();
-		for _ in 0..N {
-			array.push(T::decode(input)?);
-		}
-
-		match array.into_inner() {
-			Ok(a) => Ok(a),
-			Err(_) => panic!("We decode `N` elements; qed"),
-		}
+		decode_array(input)
 	}
 }
 
@@ -1348,6 +1446,19 @@ mod tests {
 		assert_eq!(<Vec<OptionBool>>::decode(&mut &encoded[..]).unwrap(), value);
 	}
 
+	#[cfg(feature = "bytes")]
+	#[test]
+	fn bytes_works_as_expected() {
+		let input = bytes::Bytes::from_static(b"hello");
+		let encoded = Encode::encode(&input);
+		let encoded_vec = input.to_vec().encode();
+		assert_eq!(encoded, encoded_vec);
+
+		assert_eq!(
+			&b"hello"[..],
+			bytes::Bytes::decode(&mut &encoded[..]).unwrap(),
+		);
+	}
 	fn test_encode_length<T: Encode + Decode + DecodeLength>(thing: &T, len: usize) {
 		assert_eq!(<T as DecodeLength>::len(&mut &thing.encode()[..]).unwrap(), len);
 	}
@@ -1647,6 +1758,25 @@ mod tests {
 		<[u32; 0]>::decode(&mut &encoded[..]).unwrap();
 	}
 
+
+	macro_rules! test_array_encode_and_decode {
+		( $( $name:ty ),* $(,)? ) => {
+			$(
+				paste::item! {
+					#[test]
+					fn [<test_array_encode_and_decode _ $name>]() {
+						let data: [$name; 32] = [123; 32];
+						let encoded = data.encode();
+						let decoded: [$name; 32] = Decode::decode(&mut &encoded[..]).unwrap();
+						assert_eq!(decoded, data);
+					}
+				}
+			)*
+		}
+	}
+
+	test_array_encode_and_decode!(u8, i8, u16, i16, u32, i32, u64, i64, u128, i128);
+
 	fn test_encoded_size(val: impl Encode) {
 		let length = val.using_encoded(|v| v.len());
 
